#ifndef GFXARRAY_INCLUDED // -*- C++ -*-
#define GFXARRAY_INCLUDED
#if !defined(__GNUC__)
#  pragma once
#endif

/************************************************************************

  These array classes provide typed access to a contiguous block of
  elements.  Unlike C++ arrays (e.g., int x[3];), their dimension can
  be declared at run time, and they can be resized.  The available
  array types are:

     (1)  array<T>  -- 1-D array, as in T x[];
     (2)  array2<T> -- 2-D array, as in T x[][];
     (3)  array3<T> -- 3-D array, as in T x[][][];
     (4)  varray<T> -- automatically resizing array, a lightweight
                       relative of the STL std::vector class
  
  $Id: array.h,v 1.12 2001/05/30 15:57:31 garland Exp $

 ************************************************************************/

#include "gfx.h"

// This trivial allocator is a trick we use to resize the array
// without destructing and reconstructing all the elements.
//
enum _array_alloc_policy { ARRAY_ALLOC_INPLACE };
inline void *operator new(size_t, void *p, _array_alloc_policy) { return p; }

#if _MSC_VER >= 1200
//
// This matching delete operator is necessary to avoid warnings in
// VC++ 6.0.  For some reason, it seems to periodically cause internal
// compiler errors in some GCC 2.95.x compilers.
//
inline void operator delete(void *mem, void *spot, _array_alloc_policy) { }
#endif

template<class T>
class array
{
private:
    int N;
    T *block;

protected:
    array() { }
    void init_block(int n)
	{
	    // Allocate memory for block
	    N=n; block = (T *)malloc(sizeof(T)*n);

	    // Initialize each entry
	    for(int i=0;i<n;i++) new((void *)&block[i],ARRAY_ALLOC_INPLACE) T;
	}
    void resize_block(int n)
	{
	    T *old = block;

	    // Allocate new block
	    block = (T *)realloc(old, sizeof(T)*n);

	    // Initialize all the newly allocated entries
	    for(int i=N;i<n;i++) new((void *)&block[i],ARRAY_ALLOC_INPLACE) T;

	    N = n;
	}
    void free_block()
	{
#if defined(_MSC_VER)
	    // The Microsoft compiler has a problem with the 
	    // standard syntax below.  For some reason,
	    // expanding it into the following pointer-based
	    // version makes it work.  Don't ask me why.
	    //
	    for(int i=0; i<N; i++)  { T *p = &block[i]; p->~T(); }
	    free(block);
#else
	    // Call the relevant destructors for each element before
	    // freeing the block.  Has now effect for types like 'int'.
	    //
	    for(int i=0; i<N; i++)  block[i].~T();
	    free(block);
#endif
	}

public:
    array(int n) { init_block(n); }
    ~array() { free_block(); }

    operator const T*() const { return block; }
    operator       T*()       { return block; }
    int length() const { return N; }

#ifndef HAVE_CASTING_LIMITS
    T&       operator[](int i)       { return block[i]; }
    const T& operator[](int i) const { return block[i]; }
    operator const T*()       { return block; }
#endif

    // These parenthesized accessors are included for backwards
    // compatibility.  Their continued use is discouraged.
    //
    T&       operator()(int i)       { return (*this)[i]; }
    const T& operator()(int i) const { return (*this)[i]; }

    // Primitive methods for altering the data block
    //
    void resize(int n) { resize_block(n); }
    void bitcopy(const T *a, int n) // copy bits directly
	{ memcpy(block, a, MIN(n,N)*sizeof(T)); }
    void copy(const T *a, const int n) // copy using assignment operator
	{ for(int i=0; i<MIN(n,N); i++) block[i] = a[i]; }
    void bitcopy(const array<T>& b) { bitcopy(b, b.length()); }
    void copy(const array<T>& b) { copy(b, b.length()); }

    // Restricted STL-like interface for interoperability with
    // STL-based code.
    //
    typedef T value_type;
    typedef value_type *iterator;
    typedef value_type *const_iterator;

    int size() const { return length(); }

    iterator       begin()       { return block; }
    const_iterator begin() const { return block; }
    iterator       end()       { return begin()+size(); }
    const_iterator end() const { return begin()+size(); }
};


template<class T>
class array2 : public array<T>
{
private:
    int W, H;

protected:
    array2() { }

public:
    array2(int w, int h) : array<T>(w*h) { W=w; H=h; }

    T&       operator()(int i, int j)       { return (*this)[j*W+i]; }
    const T& operator()(int i, int j) const { return (*this)[j*W+i]; }

    int width() const { return W; }
    int height() const { return H; }

    void resize(int w, int h) { W=w; H=h; array<T>::resize_block(w*h); }
};


template<class T>
class array3 : public array<T>
{
private:
    int W, H, D;

protected:
    array3() { }

public:
    array3(int w, int h, int d) : array<T>(w*h*d) { W=w; H=h; D=d; }

    T& operator()(int i, int j, int k) { return (*this)[k*W*H + j*W+i]; }
    const T& operator()(int i,int j,int k) const {return (*this)[k*W*H+j*W+i];}

    int width() const { return W; }
    int height() const { return H; }
    int depth() const { return D; }

    void resize(int w,int h,int d) {W=w;H=h;D=d;array<T>::resize_block(w*h*d);}
};

template<class T>
class varray : public array<T>
{
private:
    int fill;

public:
    varray(int n=2) : array<T>(n) { fill=0; }

    int length() const { return fill; }
    int total_space() const { return array<T>::length(); }

    int      last_id() const { return fill-1; }
    T&       last()          { return (*this)[last_id()]; }
    const T& last() const    { return (*this)[last_id()]; }

    void room_for(int len)
    {
	if( length()<len ) resize(len);
	fill = len;
    }

    T& add()
    {
	if( length()==total_space() )  resize(total_space() * 2);
	fill++;
	return last();
    }

    void add(const T& t) { add() = t; }

    void reset() { fill = 0; }
    T& drop() { return (*this)[--fill]; }
    void drop(int d) { fill -= d; }

    void remove(int i) { (*this)[i] = (*this)[--fill]; }
    void remove_inorder(int i)
    	{ memmove(&(*this)[i], &(*this)[i+1], (--fill - i)*sizeof(T)); }

    // Restricted STL-like interface for interoperability with
    // STL-based code.  Overrides select array<> definitions and
    // introduces some additional std::vector-like methods.
    //
    int size() const { return length(); }

    iterator       end()       { return begin()+size(); }
    const_iterator end() const { return begin()+size(); }

    void push_back(const T& t) { add(t); }
};

// GFXARRAY_INCLUDED
#endif
