#ifndef GFXGEOM3D_INCLUDED
#define GFXGEOM3D_INCLUDED
#if !defined(__GNUC__)
#  pragma once
#endif

/************************************************************************

  Handy 3D geometrical primitives

  $Id: geom3d.h,v 1.3 2001/03/05 23:15:00 garland Exp $

 ************************************************************************/

#include "vec3.h"
#include "vec4.h"

template<class T>
inline TVec3<T> triangle_raw_normal(const TVec3<T>& v1,
	                     const TVec3<T>& v2, const TVec3<T>& v3)
{
    TVec3<T> a = v2-v1,  b = v3-v1;
    return a^b;
}

template<class T>
inline T triangle_area(const TVec3<T>& v1,
	               const TVec3<T>& v2, const TVec3<T>& v3)
{
    return 0.5 * norm(triangle_raw_normal(v1, v2, v3));
}

template<class T>
inline TVec3<T> triangle_normal(const TVec3<T>& v1,
	                 const TVec3<T>& v2, const TVec3<T>& v3)
{
    TVec3<T> n = triangle_raw_normal(v1, v2, v3);
    unitize(n);
    return n;
}

template<class T>
inline TVec4<T> triangle_plane(const TVec3<T>& v1,
	                const TVec3<T>& v2, const TVec3<T>& v3)
{
    TVec3<T> n = triangle_normal(v1, v2, v3);
    return TVec4<T>(n, -(n*v1));
}

template<class T>
inline TVec4<T> triangle_raw_plane(const TVec3<T>& v1,
	                    const TVec3<T>& v2, const TVec3<T>& v3)
{
    TVec3<T> n = triangle_raw_normal(v1, v2, v3);
    return TVec4<T>(n, -(n*v1));
}

template<class T>
inline T triangle_compactness(const TVec3<T>& v1,
	                      const TVec3<T>& v2, const TVec3<T>& v3)
{
    const T FOUR_ROOT3 = 6.928203230275509;
    T L1 = norm2(v2 - v1);
    T L2 = norm2(v3 - v2);
    T L3 = norm2(v1 - v3);

    return FOUR_ROOT3 * triangle_area(v1, v2, v3) / (L1+L2+L3);
}

template<class Vec, class List>
void compute_bbox(Vec& min, Vec& max, const List& items)
{
    typedef typename List::const_iterator iterator;

    if( items.size()==0 )  min = max = 0;
    else                   min = max = items[0];

    for(iterator i=items.begin(); i!=items.end(); i++)
    {
	const Vec& v = *i;
	for(int j=0; j<Vec::dim(); j++)
	{
	    if( v[j] < min[j] )  min[j] = v[j];	    
	    if( v[j] > max[j] )  max[j] = v[j];
	}
    }
}

// GFXGEOM3D_INCLUDED
#endif
