#ifndef GFXSCRIPT_INCLUDED // -*- C++ -*-
#define GFXSCRIPT_INCLUDED
#if !defined(__GNUC__)
#  pragma once
#endif

/************************************************************************

  Simple low-level file format scripting support.
    
  $Id: script.h,v 1.8 2001/05/15 16:01:51 garland Exp $

 ************************************************************************/

#include "gfx.h"

#include <vector>
#include <map>
#include <string>

class Scripting
{
////////////////////////////////////////////////////////////////////////
//
// Type definitions
//
public:
    typedef int (*handler)(const char *name, char *argline, Scripting *env);

    typedef std::map<std::string, handler> Lexicon;
    typedef std::map<std::string, Lexicon*> Dictionary;
    typedef std::vector<Lexicon *> LexStack;

////////////////////////////////////////////////////////////////////////
//
// Static (class-wide) definitions
//
private:
    static Lexicon *lex_default;
    static Dictionary *dict;

public:
    static Lexicon *lookup_lexicon(const char *name);

    static void register_command(const char *name, handler proc,
				 const char *lexicon=NULL);

////////////////////////////////////////////////////////////////////////
//
// Member variables & function definitions
//
private:
    LexStack active;

protected:
    int do_mutable_string(char *line);
    handler lookup_command(const char *op, Lexicon& lex);

public:
    Scripting();
    ~Scripting();

    handler lookup_command(const char *op);


    void push_lexicon(const char *lexname);
    void pop_lexicon();

    int do_string(const char *line);
    int do_stream(std::istream& in);
    int do_file(const char *filename);
};

enum {
    SCRIPT_OK = 0,
    SCRIPT_ERR_UNDEF = 1,
    SCRIPT_ERR_SYNTAX = 2,
    SCRIPT_ERR_UNSUPPORTED = 3,
};

extern Scripting *script_environment();

inline int script_do_string(const char *line)
    { return script_environment()->do_string(line); }

inline int script_do_file(const char *filename)
    { return script_environment()->do_file(filename); }

////////////////////////////////////////////////////////////////////////
//
// Low-level support routines
//
// These procedures provide support for manipulating the command
// strings passed to scripting procedures.
//

extern char *string_trimleft(char *line);
extern char *string_trimright(char *line);
extern char *string_shift_token(char **line);

extern int string_shift_numbers(char **line, double *x, int n);
extern int string_shift_numbers(char **line, float *x, int n);
extern int string_shift_numbers(char **line, int *x, int n);

////////////////////////////////////////////////////////////////////////
//
// "Compile-time" registration mechanism
//
// We want to allow separate code modules to register their scripting
// procedures without the main program having to call intialization
// functions for each one.  So we take advantage of C++, and use
// constructors for global "variables" to accomplish all this
// initialization implicitly.
//
// Recipe for use:
//
// 	static script_defn handlers[] = {
//	    {"sample", sample_proc},
//	};
//
//	REGISTER_COMMANDS(handlers);
//

struct script_defn { const char *name; Scripting::handler proc; };

class __script_registration
{
public:
    __script_registration(script_defn *table, int n, const char *lexname=NULL)
    {
	for(int i=0; i<n; i++)
	    Scripting::register_command(table[i].name, table[i].proc, lexname);
    }
};

#define REGISTER_COMMANDS(t) \
	static __script_registration __scripted(t, sizeof(t)/sizeof(t[0]))

#define REGISTER_COMMANDS_IN(t,N) \
	static __script_registration __scripted(t, sizeof(t)/sizeof(t[0]), N);

// GFXSCRIPT_INCLUDED
#endif
