#include <gfx/gfx.h>
#include <gfx/script.h>

#include <cctype>
#include <fstream>

// NOTE: The default environment must not be declared as a variable; it must
// be dynamically allocated with 'new'.  Because of the implicit registration
// mechanism that we use, register_command() will be called during the global
// constructor phase of the program startup.  We cannot guarantee that the
// env_default constructor would be called before all the other ones.  So we
// need to take care of allocating it ourselves.
//
static Scripting *env_default = NULL;

Scripting *script_environment()
{
    if( !env_default )
	env_default = new Scripting;

    return env_default;
}

Scripting::Lexicon *Scripting::lex_default = NULL;
Scripting::Dictionary *Scripting::dict = NULL;

Scripting::Lexicon *Scripting::lookup_lexicon(const char *name)
{
    // First, make sure that the necessary static data has been initialized.
    if(!dict)  dict = new Dictionary;
    if(!lex_default)
    {
	lex_default = new Lexicon; 
	(*dict)["default"] = lex_default;
    }
    
    if(name==NULL) return lex_default;

    Dictionary::iterator i = dict->find(name);
    return (i==dict->end() ? NULL : i->second);
}

void Scripting::register_command(const char *op, handler proc,
				 const char *lexname)
{
    Lexicon *lex = lookup_lexicon(lexname);
    if( lex )
    {
	(*lex)[op] = proc;
    }
    else
    {
	lex = new Lexicon;
	(*dict)[lexname] = lex;
	(*lex)[op] = proc;
    }
}

////////////////////////////////////////////////////////////////////////
//
//
//

Scripting::Scripting()
{
    active.push_back(lookup_lexicon(NULL));
}

Scripting::~Scripting()
{
}

Scripting::handler Scripting::lookup_command(const char *op, Lexicon& lex)
{
    Lexicon::iterator i = lex.find(op);
    return (i==lex.end() ? NULL : i->second);
}

Scripting::handler Scripting::lookup_command(const char *op)
{
    for( LexStack::iterator i = active.begin(); i!=active.end(); i++)
    {
	Lexicon& lex = **i;

	handler fn = lookup_command(op, lex);
	if( fn )
	    return fn;
    }

    return NULL;
}

void Scripting::push_lexicon(const char *lexname)
{
    Lexicon *lex = lookup_lexicon(lexname);
    if(lex)
	active.push_back(lex);
}

void Scripting::pop_lexicon()
{
    if( active.size() > 1 )
	active.pop_back();
}

////////////////////////////////////////////////////////////////////////
//
//
//

int Scripting::do_mutable_string(char *line)
{
    line = string_trimleft(line);

    // Ignore comments and whitespace-only lines
    if( !line[0] || line[0]=='#' )  return 0;

    const char *op = string_shift_token(&line);

    handler fn = lookup_command(op);
    if( !fn )
    {
	std::cerr << "!! Undefined function '" << op << "'" << std::endl;
	return SCRIPT_ERR_UNDEF;
    }
    else
	return (*fn)(op, line, this);
}

int Scripting::do_string(const char *line)
{
    char *mutable_line = strdup(line);
    int rc = do_mutable_string(mutable_line);
    free(mutable_line);
    return rc;
}

int Scripting::do_stream(std::istream& in)
{
    char line[1024];

    in.getline(line, sizeof(line));
    while( !in.eof() )
    {
	do_mutable_string(line);
	in.getline(line, sizeof(line));
    }

    return SCRIPT_OK;
}

int Scripting::do_file(const char *filename)
{
    std::ifstream in(filename);
    return do_stream(in);
}

////////////////////////////////////////////////////////////////////////
//
//
//
char *string_trimleft(char *line)
{
    if(!line)  return NULL;

    while( isspace(*line) )
	line++;

    return line;
}

char *string_trimright(char *line)
{
    if(!line)  return NULL;

    char *end = line + strlen(line) - 1;
    while( end>=line && isspace(*end) ) *(end--) = '\0';
    return line;
}

char *string_shift_token(char **line)
{
    char *tok = NULL;

    if( *line )
    {
        tok = *line;

        bool override = false;
        while( **line && (override || !isspace(**line)) )
        {
            if( **line == '"' ) override = !override;
            (*line)++;
        }

        if( **line )
        {
            **line = '\0';
            *line = string_trimleft(*line+1);
        }
        else
            *line = NULL;

        if( tok[0] == '"' )
        {
            tok++;
            tok[strlen(tok)-1] = '\0';
        }

    }

    return tok;
}

static bool shift_double(char **line, double *x)
{
    char *text = string_shift_token(line);
    if( text )
    {
        char *end;
        *x = strtod(text, &end);
        if( end != text )
            return true;
    }

    return false;
}

static bool shift_float(char **line, float *x)
{
    char *text = string_shift_token(line);
    if( text )
    {
        char *end;
        *x = strtod(text, &end);
        if( end != text )
            return true;
    }

    return false;
}


static bool shift_int(char **line, int *x)
{
    char *text = string_shift_token(line);
    if( text )
    {
        char *end;
        *x = strtol(text, &end, 0);
        if( end != text )
            return true;
    }

    return false;
}


int string_shift_numbers(char **line, double *x, int n)
{
    int got;
    for(got=0; got<n && shift_double(line, &x[got]); got++)
	;
    return got;
}

int string_shift_numbers(char **line, float *x, int n)
{
    int got;
    for(got=0; got<n && shift_float(line, &x[got]); got++)
	;
    return got;

}

int string_shift_numbers(char **line, int *x, int n)
{
    int got;
    for(got=0; got<n && shift_int(line, &x[got]); got++)
	;
    return got;

}

////////////////////////////////////////////////////////////////////////
//
//
//

static int script_include(const char *op, char *argline, Scripting *env)
{
    char *filename = string_shift_token(&argline);
    if(!filename)  return SCRIPT_ERR_SYNTAX;

    return env->do_file(filename);
}


static script_defn handlers[] =
{
    {"include", script_include},
};

REGISTER_COMMANDS(handlers);
